@@ -6,11 +6,12 @@ import logging
 import awswrangler as wr
 import boto3
 import gzip
+import pendulum
 import python_graphql_client
 import pandas as pd
 import requests
 from datetime import datetime
-from typing import Optional, Dict, List
+from typing import Optional, Dict, List, Union
 
 from requests.auth import AuthBase
 from sumatra.auth import SDKKeyAuth, CognitoJwtAuth
@@ -22,6 +23,19 @@ logger = logging.getLogger("sumatra.client")
 def parse_timestamp_columns(df, columns):
     df = df.copy()
     for col in columns:
+        times = []
+        for t in df[col]:
+            ts = "NaT"
+            try:
+                ts = (
+                    pendulum.parse(t)
+                    .astimezone(pendulum.timezone("UTC"))
+                    .to_iso8601_string()
+                )
+            except:
+                pass
+            times.append(ts)
+        df[col] = times
         df[col] = pd.to_datetime(df[col], unit="ns")
         if df[col].dt.tz is None:
             df[col] = df[col].dt.tz_localize("UTC")
@@ -33,7 +47,11 @@ def tz_convert_timestamp_columns(df):
     df = df.copy()
     for col in df.columns:
         if hasattr(df[col], "dt"):
-            df[col] = df[col].dt.tz_localize("UTC").dt.tz_convert(CONFIG.timezone)
+            try:
+                df[col] = df[col].dt.tz_localize("UTC")
+            except:
+                pass
+            df[col] = df[col].dt.tz_convert(CONFIG.timezone)
     return df
 
 
@@ -450,6 +468,276 @@ class Client:
         if "errors" in ret:
             raise Exception(ret["errors"][0]["message"])
 
+    def get_inputs_from_feed(
+        self,
+        start_ts: Optional[Union[pd.Timestamp, str]] = None,
+        end_ts: Optional[Union[pd.Timestamp, str]] = None,
+        count: Optional[int] = None,
+        event_types: Optional[List[str]] = None,
+        where: Dict[str, str] = {},
+        batch_size: int = 10000,
+    ) -> List[Dict]:
+        """
+        Return the raw input events from the Event Feed.
+
+        Fetches events in descending time order from `end_ts`. May specify `count` or `start_ts`, but not both.
+
+        Arguments:
+            start_ts: Earliest event timestamp to fetch (local client timezone). If not specified, `count` will be used instead.
+            end_ts: Latest event timestamp to fetch (local client timezone) [default: now].
+            count: Number of rows to return (if start_ts not specified) [default: 10].
+            event_types: Subset of event types to fetch. [default: all]
+            where: Dictionary of equality conditions (all must be true for a match), e.g. {"zipcode": "90210", "email_domain": "gmail.com"}.
+            batch_size: Maximum number of records to fetch per GraphQL call.
+
+        Returns:
+            List of events: [{"_id": , "_type": , "_time": , [inputs...]}] (in descending time order).
+        """
+
+        where = [{"key": k, "value": v} for k, v in where.items()]
+        if batch_size < 1 or batch_size > 10000:
+            raise Exception(f"batch size: {batch_size} is out of range [1,10000]")
+        end_ts = end_ts or str(pendulum.now())
+        if isinstance(end_ts, pd.Timestamp):
+            end_ts = str(end_ts)
+        end_ts = pendulum.parse(end_ts, tz=CONFIG.timezone).astimezone(
+            pendulum.timezone("UTC")
+        )
+
+        if event_types is not None:
+            event_types = [{"type": t} for t in event_types]
+
+        query = """
+        query EventFeed($size: Int!, $end: DateTime!, $eventTypes: [EventSelection], $where: [FeatureFilter]!) {
+            events {
+                feed(
+                    from: 0
+                    size: $size
+                    end: $end
+                    types: $eventTypes
+                    where: $where
+                ) {
+                id
+                type
+                time
+                input
+                }
+            }
+        }
+        """
+        if start_ts:
+            if count:
+                raise Exception("specify only one of: start_ts or count")
+            if isinstance(start_ts, pd.Timestamp):
+                start_ts = str(start_ts)
+            start_ts = pendulum.parse(start_ts, tz=CONFIG.timezone).astimezone(
+                pendulum.timezone("UTC")
+            )
+
+            rows = []
+            done = False
+            while not done:
+                variables = {
+                    "size": batch_size,
+                    "end": end_ts.to_iso8601_string(),
+                    "where": where,
+                }
+                if event_types:
+                    variables["eventTypes"] = event_types
+                ret = self._gql_client.execute(
+                    query=query,
+                    variables=variables,
+                )
+                if "errors" in ret:
+                    raise Exception(ret["errors"][0]["message"])
+                new_rows = []
+                for event in ret["data"]["events"]["feed"]:
+                    event_time = pendulum.parse(event["time"])
+                    if event_time < start_ts:
+                        done = True
+                        break
+                    row = {
+                        "_id": event["id"],
+                        "_type": event["type"],
+                        "_time": str(event_time),
+                    }
+                    row.update(event["input"])
+                    new_rows.append(row)
+                rows.extend(new_rows)
+                if done or not new_rows:
+                    break
+                end_ts = event_time
+            return rows
+        else:  # count
+            if count is None:
+                count = 10
+            from_ = 0
+            rows = []
+            while True:
+                size = min(batch_size, count - from_)
+                if size <= 0:
+                    break
+                variables = {
+                    "size": size,
+                    "end": end_ts.to_iso8601_string(),
+                    "where": where,
+                }
+                if event_types:
+                    variables["eventTypes"] = event_types
+                ret = self._gql_client.execute(
+                    query=query,
+                    variables=variables,
+                )
+                if "errors" in ret:
+                    raise Exception(ret["errors"][0]["message"])
+                new_rows = []
+                for event in ret["data"]["events"]["feed"]:
+                    event_time = pendulum.parse(event["time"])
+                    row = {
+                        "_id": event["id"],
+                        "_type": event["type"],
+                        "_time": str(event_time),
+                    }
+                    row.update(event["input"])
+                    new_rows.append(row)
+                rows.extend(new_rows)
+                from_ += size
+                if from_ >= count:
+                    break
+                end_ts = event_time
+            return rows
+
+    def get_features_from_feed(
+        self,
+        event_type: str,
+        start_ts: Optional[Union[pd.Timestamp, str]] = None,
+        end_ts: Optional[Union[pd.Timestamp, str]] = None,
+        count: Optional[int] = None,
+        where: Dict[str, str] = {},
+        batch_size: int = 10000,
+    ) -> pd.DataFrame:
+        """
+        For a given event type, return the feature values as they were
+        calculated at event time.
+
+        Fetches events in descending time order from `end_ts`. May specify `count` or `start_ts`, but not both.
+
+        Arguments:
+            event_type: Event type name.
+            start_ts: Earliest event timestamp to fetch (local client timezone). If not specified, `count` will be used instead.
+            end_ts: Latest event timestamp to fetch (local client timezone) [default: now].
+            count: Number of rows to return (if start_ts not specified) [default: 10].
+            where: Dictionary of equality conditions (all must be true for a match), e.g. {"zipcode": "90210", "email_domain": "gmail.com"}.
+            batch_size: Maximum number of records to fetch per GraphQL call.
+
+        Returns:
+            Dataframe: _id, _time, [features...] (in descending time order).
+        """
+
+        where = [{"key": k, "value": v} for k, v in where.items()]
+        if batch_size < 1 or batch_size > 10000:
+            raise Exception(f"batch size: {batch_size} is out of range [1,10000]")
+        end_ts = end_ts or str(pendulum.now())
+        if isinstance(end_ts, pd.Timestamp):
+            end_ts = str(end_ts)
+        end_ts = pendulum.parse(end_ts, tz=CONFIG.timezone).astimezone(
+            pendulum.timezone("UTC")
+        )
+
+        query = """
+        query EventFeed($size: Int!, $end: DateTime!, $eventType: String!, $where: [FeatureFilter]!) {
+            events {
+                feed(
+                    from: 0
+                    size: $size
+                    end: $end
+                    types: [{ type: $eventType, where: $where }]
+                ) {
+                id
+                time
+                features
+                }
+            }
+        }
+        """
+        if start_ts:
+            if count:
+                raise Exception("specify only one of: start_ts or count")
+            if isinstance(start_ts, pd.Timestamp):
+                start_ts = str(start_ts)
+            start_ts = pendulum.parse(start_ts, tz=CONFIG.timezone).astimezone(
+                pendulum.timezone("UTC")
+            )
+
+            rows = []
+            done = False
+            while not done:
+                ret = self._gql_client.execute(
+                    query=query,
+                    variables={
+                        "size": batch_size,
+                        "end": end_ts.to_iso8601_string(),
+                        "eventType": event_type,
+                        "where": where,
+                    },
+                )
+                if "errors" in ret:
+                    raise Exception(ret["errors"][0]["message"])
+                new_rows = []
+                for event in ret["data"]["events"]["feed"]:
+                    event_time = pendulum.parse(event["time"])
+                    if event_time < start_ts:
+                        done = True
+                        break
+                    row = {"_id": event["id"], "_time": event_time}
+                    row.update(event["features"])
+                    new_rows.append(row)
+                rows.extend(new_rows)
+                if done or not new_rows:
+                    break
+                end_ts = event_time
+            if not rows:
+                return pd.DataFrame(columns=["_id", "_time"])
+            df = pd.DataFrame(rows)
+            df = tz_convert_timestamp_columns(df)
+            return df.set_index("_id")
+        else:  # count
+            if count is None:
+                count = 10
+            from_ = 0
+            rows = []
+            while True:
+                size = min(batch_size, count - from_)
+                if size <= 0:
+                    break
+                ret = self._gql_client.execute(
+                    query=query,
+                    variables={
+                        "size": size,
+                        "end": end_ts.to_iso8601_string(),
+                        "eventType": event_type,
+                        "where": where,
+                    },
+                )
+                if "errors" in ret:
+                    raise Exception(ret["errors"][0]["message"])
+                new_rows = []
+                for event in ret["data"]["events"]["feed"]:
+                    event_time = pendulum.parse(event["time"])
+                    row = {"_id": event["id"], "_time": event_time}
+                    row.update(event["features"])
+                    new_rows.append(row)
+                rows.extend(new_rows)
+                from_ += size
+                if from_ >= count:
+                    break
+                end_ts = event_time
+            if not rows:
+                return pd.DataFrame(columns=["_id", "_time"])
+            df = pd.DataFrame(rows)
+            df = tz_convert_timestamp_columns(df)
+            return df.set_index("_id")
+
     def get_timelines(self) -> pd.DataFrame:
         """
         Return all timelines and their metadata.
@@ -511,7 +799,7 @@ class Client:
         Return metadata about the timeline.
 
         Arguments:
-            timeline: Name of timeline.
+            timeline: Timeline name.
 
         Returns:
             Timeline metadata.
@@ -527,7 +815,7 @@ class Client:
         Delete timeline
 
         Arguments:
-            timeline: Name of timeline.
+            timeline: Timeline name.
         """
         logger.info(f"Deleting timeline '{timeline}'.")
         query = """
@@ -558,7 +846,7 @@ class Client:
         ```
 
         Arguments:
-            timeline: Name of timeline.
+            timeline: Timeline name.
 
         Returns:
             Scowl source code as string.
@@ -575,43 +863,95 @@ class Client:
             raise Exception(ret["errors"][0]["message"])
         return ret["data"]["timeline"]["scowl"]
 
+    def create_timeline_from_feed(
+        self,
+        timeline: str,
+        start_ts: Union[pd.Timestamp, str],
+        end_ts: Union[pd.Timestamp, str],
+        event_types: Optional[List[str]] = None,
+    ) -> None:
+        """
+        Create (or overwrite) timeline from the Event Feed
+
+        Arguments:
+            timeline: Timeline name.
+            start_ts: Earliest event timestamp to fetch (local client timezone).
+            end_ts: Latest event timestamp to fetch (local client timezone).
+            event_types: Event types to include (default: all).
+        """
+
+        query = """
+            mutation SaveTimeline($id: String!, $parameters: [KeyValueInput]!) {
+              saveTimeline(id: $id, source: "es", state: "processing", parameters: $parameters) {
+                id
+              }
+            }
+        """
+        if isinstance(start_ts, pd.Timestamp):
+            start_ts = str(start_ts)
+        if isinstance(end_ts, pd.Timestamp):
+            end_ts = str(end_ts)
+        start_ts = pendulum.parse(start_ts, tz=CONFIG.timezone).astimezone(
+            pendulum.timezone("UTC")
+        )
+        end_ts = pendulum.parse(end_ts, tz=CONFIG.timezone).astimezone(
+            pendulum.timezone("UTC")
+        )
+        start_str = start_ts.to_iso8601_string()
+        end_str = end_ts.to_iso8601_string()
+        parameters = [
+            {"key": "start", "value": start_str},
+            {"key": "end", "value": end_str},
+        ]
+
+        if event_types:
+            events = ",".join(event_types)
+            parameters.append({"key": "events", "value": events})
+
+        ret = self._gql_client.execute(
+            query=query, variables={"id": timeline, "parameters": parameters}
+        )
+        if "errors" in ret:
+            raise Exception(ret["errors"][0]["message"])
+        self._wait_for_timeline_processing(timeline)
+
     def create_timeline_from_dataframes(
         self,
-        df_dict: Dict[str, pd.DataFrame],
         timeline: str,
+        df_dict: Dict[str, pd.DataFrame],
         timestamp_column: Optional[str] = None,
     ) -> None:
         """
         Create (or overwrite) timeline from a collection of DataFrames—one per event type.
 
         Arguments:
-            df_dict: Dictionary from event type name to DataFrame of events.
             timeline: Timeline name.
+            df_dict: Dictionary from event type name to DataFrame of events.
             timestamp_column: Name of timestamp column, default `_time`. If column not present, current timestamp used for all events.
         """
         jsonl = ""
         for event_type, df in df_dict.items():
             jsonl += self._df_to_jsonl(df, event_type, timestamp_column)
-        self.create_timeline_from_jsonl(jsonl, timeline)
+        self.create_timeline_from_jsonl(timeline, jsonl)
 
-    def create_timeline_from_jsonl(self, jsonl: str, timeline: str) -> None:
+    def create_timeline_from_jsonl(self, timeline: str, jsonl: str) -> None:
         """
         Create (or overwrite) timeline from JSON events passed in as a string.
 
         Arguments:
-            jsonl: JSON event data, one JSON dict per line.
             timeline: Timeline name.
+            jsonl: JSON event data, one JSON dict per line.
         """
 
         if not jsonl.endswith("\n"):
             jsonl += "\n"
         data = gzip.compress(bytes(jsonl, "utf-8"))
-        self._create_timeline_from_jsonl_gz(data, timeline)
+        self._create_timeline_from_jsonl_gz(timeline, data)
 
     def _create_timeline_from_jsonl_gz(
         self,
-        data: bytes,
         timeline: str,
+        data: bytes,
     ) -> None:
         query = """
             mutation SaveTimelineMutation($id: String!,
@@ -647,41 +987,41 @@ class Client:
         if "errors" in ret:
             raise Exception(ret["errors"][0]["message"])
 
-        retries = 60
-        delay = 5.0
+        self._wait_for_timeline_processing(timeline)
+
+    def _wait_for_timeline_processing(self, timeline: str) -> None:
+        RETRIES = 180
+        DELAY = 5.0
         retry_count = 0
-        while retry_count < retries:
+        while retry_count < RETRIES:
             tl = self.get_timeline(timeline)
-            if tl.status == "processing":
-                time.sleep(delay)
-                retry_count += 1
-            else:
+            if tl.status != "processing":
                 return
+            time.sleep(DELAY)
+            retry_count += 1
+        if self.status == "processing":
+            raise Exception(f"Timed out after {DELAY * RETRIES} seconds")
 
-        raise Exception(
-            f"Timeline processing timed out after {retries * delay} seconds."
-        )
-
-    def create_timeline_from_file(self, filename: str, timeline: str) -> None:
+    def create_timeline_from_file(self, timeline: str, filename: str) -> None:
         """
         Create (or overwrite) timeline from events stored in a file.
 
         Supported file types: `.jsonl`, `.jsonl.gz`
 
         Arguments:
-            filename: Name of events file to upload.
             timeline: Timeline name.
+            filename: Name of events file to upload.
         """
 
         _, ext = _splitext(filename)
 
         if ext in (".jsonl.gz", ".json.gz"):
             with open(filename, "rb") as f:
-                self._create_timeline_from_jsonl_gz(f.read(), timeline)
+                self._create_timeline_from_jsonl_gz(timeline, f.read())
         elif ext in (".jsonl", ".json"):
             with open(filename, "r") as f:
                 jsonl = f.read()
-                self.create_timeline_from_jsonl(jsonl, timeline)
+                self.create_timeline_from_jsonl(timeline, jsonl)
         else:
             raise Exception(f"Unsupported file extension: {ext}")
 
@@ -885,7 +1225,7 @@ class Materialization:
         return ret["data"]["materialization"]
 
     def _wait_for_processing(self):
-        RETRIES = 60
+        RETRIES = 180
         DELAY = 5.0
         retry_count = 0
         while retry_count < RETRIES:
